#include <stdio.h> 
#include <math.h> 
#include <stdlib.h> 

#include <cuda.h> 
#include <cutil.h> 

#include <SCRatioCUDA_Kernel.cu> 


int2** contours_d = NULL ;
unsigned char** visit_maps_d = NULL ;
cudaArray** silhouettes_d = NULL ;
cudaChannelFormatDesc chanDesc ;

unsigned int cont_mem_size ;
unsigned int img_mem_size ; 
unsigned int map_mem_size ;
unsigned int __num_of_views__ = 0 ;
bool memory_allocated = false ;
bool device_initialized = false ;


extern "C" void initialize_device(int argc, char* argv[]) 
{
	CUT_DEVICE_INIT(argc, argv) ;

	device_initialized = true ;
}

extern "C" void 
initialize_kernel_memory(unsigned int width, unsigned int height, unsigned int n_views) 
{

	if(!device_initialized) 
	{
		printf("Device is not initialized.\n") ;
		return ;
	}

	unsigned int i = 0 ;

	__num_of_views__ = n_views ;

	// Contour & map memory allocation on the device 
	cont_mem_size = sizeof(int2) * width * height ;
	map_mem_size = sizeof(unsigned char) * width * height ;
	CUDA_SAFE_CALL(cudaMalloc((void**)&contours_d, sizeof(int2*)*n_views)) ;
	CUDA_SAFE_CALL(cudaMalloc((void**)&visit_maps_d, sizeof(unsigned char)*n_views)) ;
	for(i=0; i<n_views; i++) 
	{
		CUDA_SAFE_CALL(cudaMalloc((void**)&contours_d[i], cont_mem_size)) ;
		CUDA_SAFE_CALL(cudaMalloc((void**)&visit_maps_d[i], map_mem_size)) ;
	}



	// Silhouette memory allocation 
	img_mem_size = width * height * sizeof(unsigned char) ;
	CUDA_SAFE_CALL(cudaMalloc((void**)&silhouettes_d, sizeof(cudaArray*)*n_views)) ;
	//CUDA_SAFE_CALL(cudaMalloc((void**)&texRefPtrs, sizeof(textureReference*)*n_views)) ;
	chanDesc = cudaCreateChannelDesc(8, 0, 0, 0, cudaChannelFormatKindUnsigned);
    
	for(i=0; i<n_views; i++) 
	{
		// memory allocation 
		CUDA_SAFE_CALL(cudaMallocArray(&silhouettes_d[i], &chanDesc, width, height) ); 

		// texture initialization 
		tex0.addressMode[0] = cudaAddressModeClamp;
		tex0.addressMode[1] = cudaAddressModeClamp;
		tex0.filterMode = cudaFilterModePoint;
		tex0.normalized = false;   		
	}


	memory_allocated = true ;
}


extern "C" void release_kernel_memory() 
{
	// Release memory 
	unsigned int i=0;

	for(i=0; i<__num_of_views__; i++) 
	{
		CUDA_SAFE_CALL(cudaFree(visit_maps_d[i])) ;
		CUDA_SAFE_CALL(cudaFree(contours_d[i])) ;

		CUDA_SAFE_CALL(cudaFreeArray(silhouettes_d[i])) ;
	}

	CUDA_SAFE_CALL(cudaFree(contours_d)) ;
	CUDA_SAFE_CALL(cudaFree(visit_maps_d)) ;
	CUDA_SAFE_CALL(cudaFree(silhouettes_d)) ;
}


extern "C" void 
run_kernel(unsigned char** silhouettes, unsigned int width, unsigned int height, unsigned int n_views, unsigned char* dst ) 
{
	if(!memory_allocated) 
	{
		printf("Memory is not allocated. Allocate memory using 'initialize_kernel_memory()' before calling run_kernel(). \n") ;
		return ;
	}

	unsigned int i = 0 ;
	for(i=0; i<__num_of_views__; i++) 
	{
		// Initialize contours 
		CUDA_SAFE_CALL(cudaMemset(contours_d[i], 0, cont_mem_size)) ;
		CUDA_SAFE_CALL(cudaMemset(visit_maps_d[i], 0, map_mem_size)) ;

		// Copy silhouette 
		CUDA_SAFE_CALL(cudaMemcpyToArray(silhouettes_d[i], 0, 0, silhouettes[i], img_mem_size, 
			                                                          cudaMemcpyHostToDevice));

		// bind to texture 
		switch(i) 
		{
		case 0: CUDA_SAFE_CALL(cudaBindTextureToArray(tex0, silhouettes_d[i], chanDesc)) ; break ;
		case 1: CUDA_SAFE_CALL(cudaBindTextureToArray(tex1, silhouettes_d[i], chanDesc)) ; break ;
		case 2: CUDA_SAFE_CALL(cudaBindTextureToArray(tex2, silhouettes_d[i], chanDesc)) ; break ;
		case 3: CUDA_SAFE_CALL(cudaBindTextureToArray(tex3, silhouettes_d[i], chanDesc)) ; break ;
		case 4: CUDA_SAFE_CALL(cudaBindTextureToArray(tex4, silhouettes_d[i], chanDesc)) ; break ;
		case 5: CUDA_SAFE_CALL(cudaBindTextureToArray(tex5, silhouettes_d[i], chanDesc)) ; break ;
		case 6: CUDA_SAFE_CALL(cudaBindTextureToArray(tex6, silhouettes_d[i], chanDesc)) ; break ;
		case 7: CUDA_SAFE_CALL(cudaBindTextureToArray(tex7, silhouettes_d[i], chanDesc)) ; break ;
		}
	}

	// Call contour detection kernel 
	dim3 contour_kernel_grid_dim(1,1) ;
	dim3 contour_kernel_block_dim(__num_of_views__, 1) ;

	contourDetect<<<contour_kernel_grid_dim, contour_kernel_block_dim>>>(contours_d, visit_maps_d, __num_of_views__, width, height) ;
    
	CUT_CHECK_ERROR("Kernel execution failed");


	int2* result = (int2*)malloc(sizeof(int2) * width * height) ;
	CUDA_SAFE_CALL(cudaMemcpy(result, contours_d[0], sizeof(int2)*width*height, cudaMemcpyDeviceToHost)) ;


	for(unsigned int y=0; y<height*width; y++) 
	{
		if(result[y].x ==-100) 
			break ;

		int ind2 = result[y].x + result[y].y * width ;
		//printf("(%d %d)", result[y].x , result[y].y) ;
		dst[ind2] = 255 ;
	}

	//FILE* file = fopen("d:/result_.txt", "w") ;
	//for(i=0; i<height; i++) 
	//for(unsigned int j=0; j<width; j++) 
	//{
	//	unsigned int idx = j + i*width ;
	//	fprintf(file, "(%d, %d) : (%d, %d)\n", j, i, result[idx].x, result[idx].y) ;
	//}
	//fclose(file) ;

	free(result) ;

	return ;
}
